@@ -8,6 +8,7 @@
 //
 #include <algorithm>
 #include <stack>
+#include <cmath>
 #include <boost/algorithm/string/join.hpp>
 #include <boost/range/adaptor/reversed.hpp>
 #include <rime/composition.h>
@@ -20,13 +21,13 @@
 #include <rime/algo/syllabifier.h>
 #include <rime/dict/corrector.h>
 #include <rime/dict/dictionary.h>
+#include <rime/dict/user_dictionary.h>
 #include <rime/gear/poet.h>
 #include <rime/gear/script_translator.h>
 #include <rime/gear/translator_commons.h>
 
-
-//static const char* quote_left = "\xef\xbc\x88";
-//static const char* quote_right = "\xef\xbc\x89";
+// static const char* quote_left = "\xef\xbc\x88";
+// static const char* quote_right = "\xef\xbc\x89";
 
 namespace rime {
 
@@ -36,13 +37,17 @@ struct SyllabifyTask {
   const Code& code;
   const SyllableGraph& graph;
   size_t target_pos;
-  function<void (SyllabifyTask* task, size_t depth,
-                      size_t current_pos, size_t next_pos)> push;
-  function<void (SyllabifyTask* task, size_t depth)> pop;
+  function<void(SyllabifyTask* task,
+                size_t depth,
+                size_t current_pos,
+                size_t next_pos)>
+      push;
+  function<void(SyllabifyTask* task, size_t depth)> pop;
 };
 
 static bool syllabify_dfs(SyllabifyTask* task,
-                          size_t depth, size_t current_pos) {
+                          size_t depth,
+                          size_t current_pos) {
   if (depth == task->code.size()) {
     return current_pos == task->target_pos;
   }
@@ -74,7 +79,9 @@ class ScriptSyllabifier : public PhraseSyllabifier {
                     Corrector* corrector,
                     const string& input,
                     size_t start)
-      : translator_(translator), input_(input), start_(start),
+      : translator_(translator),
+        input_(input),
+        start_(start),
         syllabifier_(translator->delimiters(),
                      translator->enable_completion(),
                      translator->strict_spelling()) {
@@ -105,12 +112,14 @@ class ScriptTranslation : public Translation {
                     Corrector* corrector,
                     Poet* poet,
                     const string& input,
-                    size_t start)
+                    size_t start,
+                    size_t end_of_input)
       : translator_(translator),
         poet_(poet),
         start_(start),
-        syllabifier_(New<ScriptSyllabifier>(
-            translator, corrector, input, start)),
+        end_of_input_(end_of_input),
+        syllabifier_(
+            New<ScriptSyllabifier>(translator, corrector, input, start)),
         enable_correction_(corrector) {
     set_exhausted(true);
   }
@@ -121,13 +130,16 @@ class ScriptTranslation : public Translation {
  protected:
   bool CheckEmpty();
   bool IsNormalSpelling() const;
-  an<Sentence> MakeSentence(Dictionary* dict,
-                                    UserDictionary* user_dict);
-  void PrepareCandidate();
+  bool PrepareCandidate();
+  template <class QueryResult>
+  void EnrollEntries(map<int, DictEntryList>& entries_by_end_pos,
+                     const an<QueryResult>& query_result);
+  an<Sentence> MakeSentence(Dictionary* dict, UserDictionary* user_dict);
 
   ScriptTranslator* translator_;
   Poet* poet_;
   size_t start_;
+  size_t end_of_input_;
   an<ScriptSyllabifier> syllabifier_;
 
   an<DictEntryCollector> phrase_;
@@ -135,12 +147,19 @@ class ScriptTranslation : public Translation {
   an<Sentence> sentence_;
 
   an<Phrase> candidate_ = nullptr;
+  size_t candidate_index_ = 0;
+  enum CandidateSource {
+    kUninitialized,
+    kUserPhrase,
+    kSysPhrase,
+    kSentence,
+  };
+  CandidateSource candidate_source_ = kUninitialized;
 
   DictEntryCollector::reverse_iterator phrase_iter_;
   UserDictEntryCollector::reverse_iterator user_phrase_iter_;
-  size_t user_phrase_index_ = 0;
 
-  size_t max_corrections_ = 4;
+  const size_t max_corrections_ = 4;
   size_t correction_count_ = 0;
 
   bool enable_correction_;
@@ -149,9 +168,7 @@ class ScriptTranslation : public Translation {
 // ScriptTranslator implementation
 
 ScriptTranslator::ScriptTranslator(const Ticket& ticket)
-    : Translator(ticket),
-      Memory(ticket),
-      TranslatorOptions(ticket) {
+    : Translator(ticket), Memory(ticket), TranslatorOptions(ticket) {
   if (!engine_)
     return;
   if (Config* config = engine_->schema()->config()) {
@@ -175,23 +192,20 @@ an<Translation> ScriptTranslator::Query(const string& input,
     return nullptr;
   if (!segment.HasTag(tag_))
     return nullptr;
-  DLOG(INFO) << "input = '" << input
-             << "', [" << segment.start << ", " << segment.end << ")";
+  DLOG(INFO) << "input = '" << input << "', [" << segment.start << ", "
+             << segment.end << ")";
 
   FinishSession();
 
-  bool enable_user_dict = user_dict_ && user_dict_->loaded() &&
-      !IsUserDictDisabledFor(input);
+  bool enable_user_dict =
+      user_dict_ && user_dict_->loaded() && !IsUserDictDisabledFor(input);
 
+  size_t end_of_input = engine_->context()->input().length();
   // the translator should survive translations it creates
-  auto result = New<ScriptTranslation>(this,
-                                       corrector_.get(),
-                                       poet_.get(),
-                                       input,
-                                       segment.start);
-  if (!result ||
-      !result->Evaluate(dict_.get(),
-                        enable_user_dict ? user_dict_.get() : NULL)) {
+  auto result = New<ScriptTranslation>(this, corrector_.get(), poet_.get(),
+                                       input, segment.start, end_of_input);
+  if (!result || !result->Evaluate(
+                     dict_.get(), enable_user_dict ? user_dict_.get() : NULL)) {
     return nullptr;
   }
   auto deduped = New<DistinctTranslation>(result);
@@ -212,16 +226,15 @@ string ScriptTranslator::Spell(const Code& code) {
   vector<string> syllables;
   if (!dict_ || !dict_->Decode(code, &syllables) || syllables.empty())
     return result;
-  result =  boost::algorithm::join(syllables,
-                                   string(1, delimiters_.at(0)));
+  result = boost::algorithm::join(syllables, string(1, delimiters_.at(0)));
   comment_formatter_.Apply(&result);
   return result;
 }
 
 string ScriptTranslator::GetPrecedingText(size_t start) const {
-  return !contextual_suggestions_ ? string() :
-      start > 0 ? engine_->context()->composition().GetTextBefore(start) :
-      engine_->context()->commit_history().latest_text();
+  return !contextual_suggestions_ ? string()
+         : start > 0 ? engine_->context()->composition().GetTextBefore(start)
+                     : engine_->context()->commit_history().latest_text();
 }
 
 bool ScriptTranslator::Memorize(const CommitEntry& commit_entry) {
@@ -252,17 +265,10 @@ Spans ScriptSyllabifier::Syllabify(const Phrase* phrase) {
   vector<size_t> vertices;
   vertices.push_back(start_);
   SyllabifyTask task{
-    phrase->code(),
-    syllable_graph_,
-    phrase->end() - start_,
-    [&](SyllabifyTask* task, size_t depth,
-        size_t current_pos, size_t next_pos) {
-      vertices.push_back(start_ + next_pos);
-    },
-    [&](SyllabifyTask* task, size_t depth) {
-      vertices.pop_back();
-    }
-  };
+      phrase->code(), syllable_graph_, phrase->end() - start_,
+      [&](SyllabifyTask* task, size_t depth, size_t current_pos,
+          size_t next_pos) { vertices.push_back(start_ + next_pos); },
+      [&](SyllabifyTask* task, size_t depth) { vertices.pop_back(); }};
   if (syllabify_dfs(&task, 0, phrase->start() - start_)) {
     result.set_vertices(std::move(vertices));
   }
@@ -270,40 +276,36 @@ Spans ScriptSyllabifier::Syllabify(const Phrase* phrase) {
 }
 
 size_t ScriptSyllabifier::BuildSyllableGraph(Prism& prism) {
-  return (size_t)syllabifier_.BuildSyllableGraph(input_,
-                                                 prism,
+  return (size_t)syllabifier_.BuildSyllableGraph(input_, prism,
                                                  &syllable_graph_);
 }
 
-bool ScriptSyllabifier::IsCandidateCorrection(const rime::Phrase &cand) const {
+bool ScriptSyllabifier::IsCandidateCorrection(const rime::Phrase& cand) const {
   std::stack<bool> results;
-  // Perform DFS on syllable graph to find whether this candidate is a correction
-  SyllabifyTask task {
-    cand.code(),
-    syllable_graph_,
-    cand.end() - start_,
-    [&](SyllabifyTask* task, size_t depth,
-        size_t current_pos, size_t next_pos) {
-      auto id = cand.code()[depth];
-      auto it_s = syllable_graph_.edges.find(current_pos);
-      // C++ prohibit operator [] of const map
-      // if (syllable_graph_.edges[current_pos][next_pos][id].type == kCorrection)
-      if (it_s != syllable_graph_.edges.end()) {
-        auto it_e = it_s->second.find(next_pos);
-        if (it_e != it_s->second.end()) {
-          auto it_type = it_e->second.find(id);
-          if (it_type != it_e->second.end()) {
-            results.push(it_type->second.is_correction);
-            return;
-          }
-        }
-      }
-      results.push(false);
-    },
-    [&](SyllabifyTask* task, size_t depth) {
-      results.pop();
-    }
-  };
+  // Perform DFS on syllable graph to find whether this candidate is a
+  // correction
+  SyllabifyTask task{cand.code(), syllable_graph_, cand.end() - start_,
+                     [&](SyllabifyTask* task, size_t depth, size_t current_pos,
+                         size_t next_pos) {
+                       auto id = cand.code()[depth];
+                       auto it_s = syllable_graph_.edges.find(current_pos);
+                       // C++ prohibit operator [] of const map
+                       // if
+                       // (syllable_graph_.edges[current_pos][next_pos][id].type
+                       // == kCorrection)
+                       if (it_s != syllable_graph_.edges.end()) {
+                         auto it_e = it_s->second.find(next_pos);
+                         if (it_e != it_s->second.end()) {
+                           auto it_type = it_e->second.find(id);
+                           if (it_type != it_e->second.end()) {
+                             results.push(it_type->second.is_correction);
+                             return;
+                           }
+                         }
+                       }
+                       results.push(false);
+                     },
+                     [&](SyllabifyTask* task, size_t depth) { results.pop(); }};
   if (syllabify_dfs(&task, 0, cand.start() - start_)) {
     for (; !results.empty(); results.pop()) {
       if (results.top())
@@ -317,29 +319,25 @@ string ScriptSyllabifier::GetPreeditString(const Phrase& cand) const {
   const auto& delimiters = translator_->delimiters();
   std::stack<size_t> lengths;
   string output;
-  SyllabifyTask task{
-    cand.code(),
-    syllable_graph_,
-    cand.end() - start_,
-    [&](SyllabifyTask* task, size_t depth,
-        size_t current_pos, size_t next_pos) {
-      size_t len = output.length();
-      if (depth > 0 && len > 0 &&
-          delimiters.find(output[len - 1]) == string::npos) {
-        output += delimiters.at(0);
-      }
-      output += input_.substr(current_pos, next_pos - current_pos);
-      lengths.push(len);
-    },
-    [&](SyllabifyTask* task, size_t depth) {
-      output.resize(lengths.top());
-      lengths.pop();
-    }
-  };
+  SyllabifyTask task{cand.matching_code(), syllable_graph_, cand.end() - start_,
+                     [&](SyllabifyTask* task, size_t depth, size_t current_pos,
+                         size_t next_pos) {
+                       size_t len = output.length();
+                       if (depth > 0 && len > 0 &&
+                           delimiters.find(output[len - 1]) == string::npos) {
+                         output += delimiters.at(0);
+                       }
+                       output +=
+                           input_.substr(current_pos, next_pos - current_pos);
+                       lengths.push(len);
+                     },
+                     [&](SyllabifyTask* task, size_t depth) {
+                       output.resize(lengths.top());
+                       lengths.pop();
+                     }};
   if (syllabify_dfs(&task, 0, cand.start() - start_)) {
     return translator_->FormatPreedit(output);
-  }
-  else {
+  } else {
     return string();
   }
 }
@@ -352,94 +350,101 @@ string ScriptSyllabifier::GetOriginalSpelling(const Phrase& cand) const {
   return string();
 }
 
+template <class Ptr, class Iter>
+static bool has_exact_match_phrase(Ptr ptr, Iter iter, size_t consumed) {
+  return ptr && iter->first == consumed && !iter->second.exhausted() &&
+         iter->second.Peek()->IsExactMatch();
+}
+
 // ScriptTranslation implementation
 
 bool ScriptTranslation::Evaluate(Dictionary* dict, UserDictionary* user_dict) {
   size_t consumed = syllabifier_->BuildSyllableGraph(*dict->prism());
   const auto& syllable_graph = syllabifier_->syllable_graph();
+  bool predict_word =
+      translator_->enable_completion() && start_ + consumed == end_of_input_;
 
-  phrase_ = dict->Lookup(syllable_graph, 0);
+  phrase_ = dict->Lookup(syllable_graph, 0, predict_word);
   if (user_dict) {
-    user_phrase_ = user_dict->Lookup(syllable_graph, 0);
+    const size_t kUnlimitedDepth = 0;
+    const size_t kNumSyllablesToPredictWord = 4;
+    user_phrase_ =
+        user_dict->Lookup(syllable_graph, 0, kUnlimitedDepth,
+                          predict_word ? kNumSyllablesToPredictWord : 0);
   }
   if (!phrase_ && !user_phrase_)
     return false;
-  // make sentences when there is no exact-matching phrase candidate
-  size_t translated_len = 0;
-  if (phrase_ && !phrase_->empty())
-    translated_len = (std::max)(translated_len, phrase_->rbegin()->first);
-  if (user_phrase_ && !user_phrase_->empty())
-    translated_len = (std::max)(translated_len, user_phrase_->rbegin()->first);
-  if (translated_len < consumed &&
-      syllable_graph.edges.size() > 1) {  // at least 2 syllables required
-    sentence_ = MakeSentence(dict, user_dict);
-  }
 
   if (phrase_)
     phrase_iter_ = phrase_->rbegin();
   if (user_phrase_)
     user_phrase_iter_ = user_phrase_->rbegin();
+
+  // make sentences when there is no exact-matching phrase candidate
+  bool has_at_least_two_syllables = syllable_graph.edges.size() >= 2;
+  if (has_at_least_two_syllables &&
+      !has_exact_match_phrase(phrase_, phrase_iter_, consumed) &&
+      !has_exact_match_phrase(user_phrase_, user_phrase_iter_, consumed)) {
+    sentence_ = MakeSentence(dict, user_dict);
+  }
+
   return !CheckEmpty();
 }
 
 bool ScriptTranslation::Next() {
-  bool is_correction;
   do {
-    is_correction = false;
     if (exhausted())
       return false;
-    if (sentence_) {
-      sentence_.reset();
-      return !CheckEmpty();
+    if (candidate_source_ == kUninitialized) {
+      PrepareCandidate();  // to determine candidate_source_
     }
-    int user_phrase_code_length = 0;
-    if (user_phrase_ && user_phrase_iter_ != user_phrase_->rend()) {
-      user_phrase_code_length = user_phrase_iter_->first;
-    }
-    int phrase_code_length = 0;
-    if (phrase_ && phrase_iter_ != phrase_->rend()) {
-      phrase_code_length = phrase_iter_->first;
-    }
-    if (user_phrase_code_length > 0 &&
-        user_phrase_code_length >= phrase_code_length) {
-      DictEntryList& entries(user_phrase_iter_->second);
-      if (++user_phrase_index_ >= entries.size()) {
-        ++user_phrase_iter_;
-        user_phrase_index_ = 0;
-      }
-    }
-    else if (phrase_code_length > 0) {
-      DictEntryIterator& iter(phrase_iter_->second);
-      if (!iter.Next()) {
-        ++phrase_iter_;
-      }
+    switch (candidate_source_) {
+      case kUninitialized:
+        break;
+      case kSentence:
+        sentence_.reset();
+        break;
+      case kUserPhrase: {
+        UserDictEntryIterator& uter(user_phrase_iter_->second);
+        if (!uter.Next()) {
+          ++user_phrase_iter_;
+        }
+      } break;
+      case kSysPhrase: {
+        DictEntryIterator& iter(phrase_iter_->second);
+        if (!iter.Next()) {
+          ++phrase_iter_;
+        }
+      } break;
     }
+    candidate_.reset();
+    candidate_source_ = kUninitialized;
     if (enable_correction_) {
-      PrepareCandidate();
-      if (!candidate_) {
+      // populate next candidate and skip it if it's a correction beyond max
+      // numbers.
+      if (!PrepareCandidate()) {
         break;
       }
-      is_correction = syllabifier_->IsCandidateCorrection(*candidate_);
     }
-  } while ( // limit the number of correction candidates
-      enable_correction_ &&
-      is_correction &&
-      correction_count_ > max_corrections_);
-  if (is_correction) {
-    ++correction_count_;
+  } while (enable_correction_ &&
+           syllabifier_->IsCandidateCorrection(*candidate_) &&
+           // limit the number of correction candidates
+           ++correction_count_ > max_corrections_);
+  if (!CheckEmpty()) {
+    ++candidate_index_;
+    return true;
   }
-  return !CheckEmpty();
+  return false;
 }
 
 bool ScriptTranslation::IsNormalSpelling() const {
   const auto& syllable_graph = syllabifier_->syllable_graph();
   return !syllable_graph.vertices.empty() &&
-      (syllable_graph.vertices.rbegin()->second == kNormalSpelling);
+         (syllable_graph.vertices.rbegin()->second == kNormalSpelling);
 }
 
 an<Candidate> ScriptTranslation::Peek() {
-  PrepareCandidate();
-  if (!candidate_) {
+  if (candidate_source_ == kUninitialized && !PrepareCandidate()) {
     return nullptr;
   }
   if (candidate_->preedit().empty()) {
@@ -447,33 +452,38 @@ an<Candidate> ScriptTranslation::Peek() {
   }
   if (candidate_->comment().empty()) {
     auto spelling = syllabifier_->GetOriginalSpelling(*candidate_);
-    if (!spelling.empty() && spelling != candidate_->preedit()) {
-      candidate_->set_comment(/*quote_left + */spelling/* + quote_right*/);
+    if (!spelling.empty() && (translator_->always_show_comments() ||
+                              spelling != candidate_->preedit())) {
+      candidate_->set_comment(/*quote_left + */ spelling /* + quote_right*/);
     }
   }
   candidate_->set_syllabifier(syllabifier_);
   return candidate_;
 }
 
-void ScriptTranslation::PrepareCandidate() {
+static bool always_true() {
+  return true;
+}
+
+template <typename T>
+inline static bool prefer_user_phrase(
+    T user_phrase_weight,
+    T sys_phrase_weight,
+    function<bool()> compare_on_tie = always_true) {
+  return user_phrase_weight > sys_phrase_weight ||
+         (user_phrase_weight == sys_phrase_weight && compare_on_tie());
+}
+
+bool ScriptTranslation::PrepareCandidate() {
   if (exhausted()) {
+    candidate_source_ = kUninitialized;
     candidate_ = nullptr;
-    return;
+    return false;
   }
   if (sentence_) {
-    if (sentence_->preedit().empty()) {
-      sentence_->set_preedit(syllabifier_->GetPreeditString(*sentence_));
-    }
-    if (sentence_->comment().empty()) {
-      auto spelling = syllabifier_->GetOriginalSpelling(*sentence_);
-      if (!spelling.empty() &&
-          (translator_->always_show_comments() ||
-              spelling != sentence_->preedit())) {
-        sentence_->set_comment(/*quote_left + */spelling/* + quote_right*/);
-      }
-    }
+    candidate_source_ = kSentence;
     candidate_ = sentence_;
-    return;
+    return true;
   }
   size_t user_phrase_code_length = 0;
   if (user_phrase_ && user_phrase_iter_ != user_phrase_->rend()) {
@@ -483,37 +493,49 @@ void ScriptTranslation::PrepareCandidate() {
   if (phrase_ && phrase_iter_ != phrase_->rend()) {
     phrase_code_length = phrase_iter_->first;
   }
-  an<Phrase> cand;
   if (user_phrase_code_length > 0 &&
-      user_phrase_code_length >= phrase_code_length) {
-    DictEntryList& entries(user_phrase_iter_->second);
-    const auto& entry(entries[user_phrase_index_]);
+      prefer_user_phrase(user_phrase_code_length, phrase_code_length, [this]() {
+        const int kNumExactMatchOnTop = 1;
+        size_t full_code_length = end_of_input_ - start_;
+        return candidate_index_ >= kNumExactMatchOnTop ||
+               prefer_user_phrase(
+                   has_exact_match_phrase(user_phrase_, user_phrase_iter_,
+                                          full_code_length),
+                   has_exact_match_phrase(phrase_, phrase_iter_,
+                                          full_code_length));
+      })) {
+    UserDictEntryIterator& uter = user_phrase_iter_->second;
+    const auto& entry = uter.Peek();
     DLOG(INFO) << "user phrase '" << entry->text
                << "', code length: " << user_phrase_code_length;
-    cand = New<Phrase>(translator_->language(),
-                       "user_phrase",
-                       start_,
-                       start_ + user_phrase_code_length,
-                       entry);
-    cand->set_quality(exp(entry->weight) +
-                      translator_->initial_quality() +
-                      (IsNormalSpelling() ? 0.5 : -0.5));
-  }
-  else if (phrase_code_length > 0) {
-    DictEntryIterator& iter(phrase_iter_->second);
-    const auto& entry(iter.Peek());
+    candidate_source_ = kUserPhrase;
+    candidate_ =
+        New<Phrase>(translator_->language(),
+                    entry->IsPredictiveMatch() ? "completion" : "user_phrase",
+                    start_, start_ + user_phrase_code_length, entry);
+    candidate_->set_quality(std::exp(entry->weight) +
+                            translator_->initial_quality() +
+                            (IsNormalSpelling() ? 0.5 : -0.5));
+    return true;
+  } else if (phrase_code_length > 0) {
+    DictEntryIterator& iter = phrase_iter_->second;
+    const auto& entry = iter.Peek();
     DLOG(INFO) << "phrase '" << entry->text
                << "', code length: " << phrase_code_length;
-    cand = New<Phrase>(translator_->language(),
-                       "phrase",
-                       start_,
-                       start_ + phrase_code_length,
-                       entry);
-    cand->set_quality(exp(entry->weight) +
-                      translator_->initial_quality() +
-                      (IsNormalSpelling() ? 0 : -1));
+    candidate_source_ = kSysPhrase;
+    candidate_ =
+        New<Phrase>(translator_->language(),
+                    entry->IsPredictiveMatch() ? "completion" : "phrase",
+                    start_, start_ + phrase_code_length, entry);
+    candidate_->set_quality(std::exp(entry->weight) +
+                            translator_->initial_quality() +
+                            (IsNormalSpelling() ? 0 : -1));
+    return true;
+  } else {
+    candidate_source_ = kUninitialized;
+    candidate_ = nullptr;
+    return false;
   }
-  candidate_ = cand;
 }
 
 bool ScriptTranslation::CheckEmpty() {
@@ -522,36 +544,41 @@ bool ScriptTranslation::CheckEmpty() {
   return exhausted();
 }
 
+template <class QueryResult>
+void ScriptTranslation::EnrollEntries(
+    map<int, DictEntryList>& entries_by_end_pos,
+    const an<QueryResult>& query_result) {
+  if (query_result) {
+    for (auto& y : *query_result) {
+      DictEntryList& homophones = entries_by_end_pos[y.first];
+      while (homophones.size() < translator_->max_homophones() &&
+             !y.second.exhausted()) {
+        homophones.push_back(y.second.Peek());
+        if (!y.second.Next())
+          break;
+      }
+    }
+  }
+}
+
 an<Sentence> ScriptTranslation::MakeSentence(Dictionary* dict,
                                              UserDictionary* user_dict) {
   const int kMaxSyllablesForUserPhraseQuery = 5;
   const auto& syllable_graph = syllabifier_->syllable_graph();
   WordGraph graph;
   for (const auto& x : syllable_graph.edges) {
-    UserDictEntryCollector& dest(graph[x.first]);
+    auto& same_start_pos = graph[x.first];
     if (user_dict) {
-      auto user_phrase = user_dict->Lookup(syllable_graph, x.first,
-                                           kMaxSyllablesForUserPhraseQuery);
-      if (user_phrase)
-        dest.swap(*user_phrase);
-    }
-    if (auto phrase = dict->Lookup(syllable_graph, x.first)) {
-      // merge lookup results
-      for (auto& y : *phrase) {
-        DictEntryList& entries(dest[y.first]);
-        while (entries.size() < translator_->max_homophones() &&
-               !y.second.exhausted()) {
-          entries.push_back(y.second.Peek());
-          if (!y.second.Next())
-            break;
-        }
-      }
+      EnrollEntries(same_start_pos,
+                    user_dict->Lookup(syllable_graph, x.first,
+                                      kMaxSyllablesForUserPhraseQuery));
     }
+    // merge lookup results
+    EnrollEntries(same_start_pos, dict->Lookup(syllable_graph, x.first));
   }
   if (auto sentence =
-      poet_->MakeSentence(graph,
-                          syllable_graph.interpreted_length,
-                          translator_->GetPrecedingText(start_))) {
+          poet_->MakeSentence(graph, syllable_graph.interpreted_length,
+                              translator_->GetPrecedingText(start_))) {
     sentence->Offset(start_);
     sentence->set_syllabifier(syllabifier_);
     return sentence;
